#!/usr/bin/env python
#
#   build-zfnann.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from itertools import izip
import random

from genann import genetics, neuralnet, util

INPUT_COUNT = 3
HIDDEN_COUNT = 8
OUTPUT_COUNT = 1

POPULATION_SIZE = 50
CROSSOVER_RATE = 0.5
MUTATION_RATE = 0.25
MAX_PERTURBATION = 0.3
ELITISM = 3
ELITE_COPIES = 1

TEST_SERVERS = 150
REDUNDANCY_TEST_SIZE = 100
REDUNDANCY_SWEET_SPOT = 65.0

class BlobSimulator(object):
    def __init__(self, uptime, data_used, total_data):
        self.uptime = uptime
        self.data_left = 1.0 - data_used / total_data

def display_generation(population, generation=None):
    min_fit, avg_fit, max_fit = get_fitness_stats(population)
    if generation is not None:
        print "Generation %i:" % generation
        indent = "  "
    else:
        indent = ""
    print indent + "Maximum Fitness:", max_fit
    print indent + "Average Fitness:", avg_fit

def get_fitness_stats(population):
    fitnesses = [fitness(genome) for genome in population]
    return min(fitnesses), sum(fitnesses) / len(fitnesses), max(fitnesses)

def generate(input_count, lengths):
    genes = []
    for i, count in enumerate(lengths):
        # Determine number of weights
        if i == 0:
            weight_count = input_count
        else:
            weight_count = lengths[i - 1]
        # Create requested number of neurons
        for j in xrange(count):
            weights = [util.random_delta() for k in xrange(weight_count)]
            bias = random.random()
            genes.append(neuralnet.NeuronGene(weights, bias))
    return neuralnet.NeuralGenome(genes)

def mutate(genome):
    new_genes = [gene.point_mutate(MUTATION_RATE, MAX_PERTURBATION)
                 for gene in genome]
    return neuralnet.NeuralGenome(new_genes)

def make_net(genome):
    return genome.to_net([HIDDEN_COUNT, OUTPUT_COUNT],
                         [neuralnet.Neuron.make_logistic_activator(),
                          neuralnet.Neuron.step_activator])

def score_relative(value, set_min, set_max):
    assert set_max >= set_min
    return (value - set_min) / (set_max - set_min)

def get_perfect_redundancy(net):
    server_count = 0
    for i in xrange(REDUNDANCY_TEST_SIZE):
        results = net.evaluate([1.0, 1.0, server_count / REDUNDANCY_TEST_SIZE])
        if results[0]:
            server_count += 1
    return (server_count / REDUNDANCY_TEST_SIZE) * 100.0

def fitness(genome):
    net = make_net(genome)
    blob_size = 4096
    selection = []
    weight = 1.0
    # Evaluate redundancy
    redundancy = get_perfect_redundancy(net)
    if redundancy < REDUNDANCY_SWEET_SPOT:
        redundancy_score = redundancy / REDUNDANCY_SWEET_SPOT
    elif redundancy > REDUNDANCY_SWEET_SPOT:
        redundancy_score = ((100.0 - redundancy) /
                            (100.0 - REDUNDANCY_SWEET_SPOT))
    else:
        redundancy_score = 1.0
    # Evaluate decisions
    for server in servers:
        redundancy = selection.count(True) / len(servers)
        results = net.evaluate([server.uptime / 100.0,
                                server.data_left,
                                redundancy,])
        chosen = bool(results[0])
        selection.append(chosen)
    # Score decisions
    server_count = selection.count(True)
    # Space left and uptime
    total_uptime = 0.0
    total_space = 0.0
    for chosen, server in izip(selection, servers):
        if chosen:
            total_uptime += server.uptime
            total_space += server.data_left
    if server_count > 0:
        uptime_score = score_relative(total_uptime / server_count,
                                      min_uptime, max_uptime)
        space_score = score_relative(total_space / server_count,
                                     min_dataleft, max_dataleft)
    else:
        uptime_score = space_score = 0.0
    return 1.0 + (redundancy_score * 6.0 +
                  uptime_score * 1.0 +
                  space_score * 3.0) * weight

# Seed random number generator
random.seed()
# Create server simulators
servers = []
for i in xrange(TEST_SERVERS):
    percent = random.randint(0, 100)
    n1 = random.randint(0, 1024 ** 2)
    n2 = random.randint(0, 1024 ** 2)
    space_used, space_total = min(n1, n2), max(n1, n2)
    servers.append(BlobSimulator(percent, space_used, space_total))
min_uptime = min(server.uptime for server in servers)
max_uptime = max(server.uptime for server in servers)
min_dataleft = min(server.data_left for server in servers)
max_dataleft = max(server.data_left for server in servers)
# Create initial population
population_zero = [generate(INPUT_COUNT, [HIDDEN_COUNT, OUTPUT_COUNT])
                   for i in xrange(POPULATION_SIZE)]
# Create more populations
generation = 0
population = population_zero
crossover_func = util.limit_function(neuralnet.NeuralGenome.two_point_crossover,
                                     CROSSOVER_RATE)
try:
    while True:
        generation += 1
        print "Calculating generation %i..." % generation
        population = genetics.population_step(
            population, fitness,
            crossover_func=crossover_func, mutate_func=mutate,
            elites=ELITISM, elite_copies=ELITE_COPIES)
        display_generation(population, generation)
except KeyboardInterrupt:
    pass
# Find best neural net
fitnesses = [fitness(genome) for genome in population]
best_index = fitnesses.index(max(fitnesses))
best_net = make_net(population[best_index])
# Display some simple tests
print "TEST RESULTS"
print "Redundancy:", get_perfect_redundancy(best_net)
print "[UP , DAT, RED]"
#               UP   DAT  RED  EXP
test_inputs = [[0.0, 0.0, 0.0, 0.0],
               [0.0, 0.0, 1.0, 0.0],
               [0.0, 1.0, 0.0, 0.0],
               [0.0, 1.0, 1.0, 0.0],
               [1.0, 0.0, 0.0, 0.0],
               [1.0, 0.0, 1.0, 0.0],
               [1.0, 1.0, 0.0, 1.0],
               [1.0, 1.0, 1.0, 0.0],
               [0.75, 1.0, 0.25, 1.0],
               [0.75, 0.5, 0.25, 1.0],
               [0.25, 0.1, 0.25, 0.0],
               [0.5, 0.5, 0.5],
               [0.75, 1.0, 0.9, 0.0],
               [0.25, 0.75, 0.9, 0.0],]
for input_set in test_inputs:
    if len(input_set) > INPUT_COUNT:
        input_set, expectation = input_set[:-1], input_set[-1]
        print "%r --> %i (wanting %i)" % \
            (input_set, best_net.evaluate(input_set)[0], expectation)
    else:
        print "%r --> %i" % (input_set, best_net.evaluate(input_set)[0])
# Display best neural net
indent = " " * 4
print "NEURAL NET SETTINGS"
print "["
for layer in best_net.layers:
    print indent + "["
    for neuron in layer:
        print indent * 2 + repr(list(neuron.weights) + [neuron.bias]) + ","
    print indent + "],"
print "]"
