#!/usr/bin/env python
#
#   zfilenet/discovery.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
High-level interface to Bonjour

:Variables:
    BLOB_SERVICE_TYPE : str
        The service type for a blob server
    INDEX_SERVICE_TYPE : str
        The service type for an index server's JSON-RPC API
    LOCAL_DOMAIN : str
        The local domain (i.e. ``local.``)
"""

from __future__ import absolute_import
import pybonjour
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IReadDescriptor
from zope import interface

from . import util

__author__ = 'Ross Light'
__date__ = 'June 12, 2008'
__all__ = ['BLOB_SERVICE_TYPE',
           'INDEX_SERVICE_TYPE',
           'LOCAL_DOMAIN',
           'Browser',
           'broadcast',
           'browse',
           'resolve',]

BLOB_SERVICE_TYPE = '_zfn-blob._tcp'
INDEX_SERVICE_TYPE = '_zfn-index._tcp'
LOCAL_DOMAIN = 'local.'

class ServiceDescriptor(object):
    """
    A Twisted wrapper for the Bonjour daemon connection.
    
    :IVariables:
        sdref : file-like
            The service descriptor file
    """
    interface.implements(IReadDescriptor)
    
    def __init__(self, sdref):
        self.sdref = sdref
    
    def doRead(self):
        pybonjour.DNSServiceProcessResult(self.sdref)
    
    def fileno(self):
        return self.sdref.fileno()
    
    def logPrefix(self):
        return "bonjour"
    
    def connectionLost(self, reason):
        pass

class Browser(object):
    """
    A browser for Bonjour services.
    
    :IVariables:
        service_type : str
            The service type being searched for
        services : list of tuple
            List of services found so far.  Each tuple contains the domain,
            the name, and the interface index.
        indefinite : bool
            Whether the browser will continue forever, or until all the
            currently available services are recorded
        progress : `util.ProgressDeferred`
            A progress deferred called every time a change is made
        finished : twisted.internet.defer.Deferred
            A deferred that calls back when the browser is finished
    """
    def __init__(self, reactor, service_type, indefinite=False):
        # Get variables
        self.reactor = reactor
        self.service_type = service_type
        self.services = []
        self.indefinite = indefinite
        # Set up deferreds
        self.progress = util.ProgressDeferred()
        self.finished = Deferred()
        # Set up browser
        self._setup_browser()
    
    def _setup_browser(self):
        self.sdref = pybonjour.DNSServiceBrowse(regtype=self.service_type,
                                                callBack=self.service_found)
        self.descriptor = ServiceDescriptor(self.sdref)
        self.reactor.addReader(self.descriptor)
    
    def service_found(self, sdref, flags, iface_index, error_code,
                      service_name, regtype, domain):
        if error_code != pybonjour.kDNSServiceErr_NoError:
            # TODO: Errback
            return
        # Update service list
        service = (domain, service_name, iface_index)
        if flags & pybonjour.kDNSServiceFlagsAdd:
            # Service found
            self.services.append(service)
        else:
            # Service lost
            self.services.remove(service)
        # Call back with progress
        self.progress.callback(self.services)
        # Check to see if we're done
        more_coming = bool(flags & pybonjour.kDNSServiceFlagsMoreComing)
        if not self.indefinite and not more_coming:
            self.stop()
    
    def _close(self):
        self.reactor.removeReader(self.descriptor)
        # TODO: Determine why this causes an error
        # self.sdref.close()
    
    def stop(self):
        """Forcefully stops the browser."""
        self._close()
        self.finished.callback(self.services)

def broadcast(reactor, service_type, port, name=None):
    """
    Announce a service over Bonjour.
    
    :Parameters:
        reactor
            The Twisted reactor to listen to Bonjour events with
        service_type : str
            The service type to broadcast (e.g. ``_foo._tcp``)
        port : int
            The port number we're broadcasting for
        name : unicode
            The human-readable name of the service.  Defaults to the host name.
    :Returns: A deferred that calls back when the broadcast is complete
    :ReturnType: twisted.internet.defer.Deferred
    """
    def _broadcast_callback(sdref, flags, error_code, name, regtype, domain):
        if error_code == pybonjour.kDNSServiceErr_NoError:
            d.callback([sdref, name, domain])
    # Create broadcaster
    d = Deferred()
    sdref = pybonjour.DNSServiceRegister(name=name,
                                         regtype=service_type,
                                         port=port,
                                         callBack=_broadcast_callback,)
    reactor.addReader(ServiceDescriptor(sdref))
    return d

def browse(reactor, service_type, indefinite=False):
    """
    Browse for services over Bonjour.
    
    :Parameters:
        reactor
            The Twisted reactor to listen to Bonjour events with
        service_type : str
            The service type to broadcast (e.g. ``_foo._tcp``)
    :Keywords:
        indefinite : bool
            Whether the browser stops after finding the available services.
    :Returns: The browser in use
    :ReturnType: Browser
    """
    return Browser(reactor, service_type, indefinite)

def resolve(reactor, service_type, domain, name, interface_index=0):
    """
    Resolves a service using Bonjour.
    
    :Parameters:
        reactor
            The Twisted reactor to listen to Bonjour events with
        service_type : str
            The service type to broadcast (e.g. ``_foo._tcp``)
        domain : str
            The domain to resolve in (usually ``local.``)
        name : unicode
            The human-readable name of the service.
        interface_index : int
            The interface to resolve with.  Should only be used with services
            found with `browse`.
    :Returns: A deferred called with a ``(host, port)`` pair when the service
              is resolved.
    :ReturnType: twisted.internet.defer.Deferred
    """
    def _resolve_callback(sdref, flags, iface_index, error_code, full_name,
                          host_target, port, txt_record):
        sdref.close()
        reactor.removeReader(descriptor)
        if error_code == pybonjour.kDNSServiceErr_NoError:
            d.callback((host_target, port))
    # Start resolver
    d = Deferred()
    sdref = pybonjour.DNSServiceResolve(0,
                                        interface_index,
                                        name,
                                        service_type,
                                        domain,
                                        _resolve_callback)
    descriptor = ServiceDescriptor(sdref)
    reactor.addReader(descriptor)
    return d
