#!/usr/bin/env python
#
#   zfilenet/indexserver/__init__.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Index server implementation."""

from __future__ import absolute_import
import logging
import sys

import progtools

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__version__ = '0.2.0b1'
__docformat__ = 'reStructuredText'
__all__ = ['blobann',
           'blobmanager',
           'dav',
           'index',
           'rpcapi',
           # Module contents
           'usage',
           'description',
           'log',
           'main',]

### PACKAGE CONTENTS ###

from . import blobann
from . import blobmanager
from . import dav
from . import index
from . import rpcapi

### MODULE CONTENTS ###

usage = "usage: %prog [-p PORT] [-n NAME] [database]"
description = "Index/WebDAV server for ZFN"
log = logging.getLogger('zfilenet.indexserver')

@progtools.reportexits
def main(args=None):
    from twisted.internet import reactor
    from .. import util
    # Parse arguments
    if args is None:
        args = sys.argv[1:]
    parser = progtools.OptionParser(usage=usage, version=__version__,
                                    description=description,)
    parser.set_defaults(rpc_port=rpcapi.DEFAULT_PORT,
                        rpc_service_name=None,
                        dav_port=dav.DEFAULT_PORT,
                        dav_service_name=None,)
    util.add_log_options(parser)
    parser.add_option('-p', '--dav-port',
                      action='store',
                      dest='dav_port',
                      type='int',
                      metavar="PORT",
                      help="Changes the port the WebDAV server runs on. "
                           "[default: %default]")
    parser.add_option('-n', '--dav-name',
                      action='store',
                      dest='dav_service_name',
                      metavar="NAME",
                      help="The WebDAV service name broadcast with Bonjour. "
                           "[default: host name]")
    parser.add_option('--rpc-port',
                      action='store',
                      dest='rpc_port',
                      type='int',
                      metavar="PORT",
                      help="Changes the port the RPC server runs on. "
                           "[default: %default]")
    parser.add_option('--rpc-name',
                      action='store',
                      dest='rpc_service_name',
                      metavar="NAME",
                      help="The RPC service name broadcast with Bonjour. "
                           "[default: host name]")
    options, args = parser.parse_args(args)
    util.setup_log_options(options)
    if len(args) == 0:
        db_file = index.DEFAULT_DB_FILE
    elif len(args) == 1:
        db_file = args[0]
    else:
        raise progtools.UsageError()
    # Connect to database
    log.debug("Using database: %s", db_file)
    try:
        engine = index.connect(db_file)
    except errors.WrongVersionError:
        log.critical("'%s' was created from an incompatible version.\n"
                     "Try using the export/import scripts to upgrade your\n"
                     "database, then run this program again.", db_file)
        return 3
    # Start blob manager
    log.info("Initializing blob tracker...")
    blob_server_list = blobmanager.BlobServerList(reactor)
    # Create servers
    rpcapi.create_server(reactor, engine, blob_server_list,
                         port=options.rpc_port,
                         service_name=options.rpc_service_name,)
    dav.create_server(reactor, engine, blob_server_list,
                      port=options.dav_port,
                      service_name=options.dav_service_name,)
    # Run reactor
    log.info("Running...")
    reactor.run()
    return 0

if __name__ == '__main__':
    sys.exit(main())
