#!/usr/bin/env python
#
#   zfilenet/indexserver/blobmanager.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Index blob managing"""

from __future__ import absolute_import, division
from base64 import b64decode, b64encode
import logging
import time

import pinet
from twisted.internet.defer import Deferred

from . import blobann
from .. import discovery, util

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__docformat__ = 'reStructuredText'
__all__ = ['log',
           'NoServersError',
           'BlobServerFactory',
           'BlobServerList',
           'BlobFetcher',
           'BlobWriter',]

log = logging.getLogger('zfilenet.indexserver.blobmanager')
_get_now = time.clock

class NoServersError(Exception):
    """
    Exception raised when `BlobServerList.write_blob` doesn't have any servers
    to write to.
    """

class BlobServerFactory(pinet.jsonrpc.JSON_RPCFactory):
    """Custom factory for blob servers."""
    def __init__(self, server_list):
        pinet.jsonrpc.JSON_RPCFactory.__init__(self)
        self.server_list = server_list
    
    def get_uptime(self):
        """
        Calculate the server's uptime relative to the program's.
        
        This is just a shortcut for `BlobServerList.get_uptime`.
        
        :Returns: The uptime as a percentage (0.0 to 1.0)
        :ReturnType: float
        """
        return self.server_list.get_uptime(self)
    
    def clientConnectionFailed(self, connector, reason):
        # Log failed connection
        address = connector.getDestination()
        log.debug("Connection failed: %s:%i [%s]",
                  address.host, address.port, reason.value)
        # Call superclass implementation
        pinet.jsonrpc.JSON_RPCFactory.clientConnectionFailed(
            self, connector, reason)
        # Drop server from list
        self.server_list.drop_server(self)
    
    def clientConnectionLost(self, connector, reason):
        # Log lost connection
        address = connector.getDestination()
        log.debug("Connection lost: %s:%i [%s]",
                  address.host, address.port, reason.value)
        # Call superclass implementation
        pinet.jsonrpc.JSON_RPCFactory.clientConnectionLost(
            self, connector, reason)
        # Drop server from list
        self.server_list.drop_server(self)

class BlobServerList(object):
    """
    Tracks all of the blob servers that are available.
    
    :IVariables:
        servers : dict
            Dictionary of ``service: server`` pairs
        history : dict
            Internal history representation
        browser : `discovery.Browser`
            Blob server browser
    """
    def __init__(self, reactor):
        self.reactor = reactor
        self.servers = {}
        # We store history information here, since factories are transient
        self.history = {}
        self._start_time = _get_now()
        self.browser = discovery.browse(self.reactor,
                                        discovery.BLOB_SERVICE_TYPE,
                                        indefinite=True)
        self.browser.progress.addCallback(self._blob_servers_changed)
    
    def _blob_servers_changed(self, services):
        new_service_set = frozenset(services)
        current_service_set = frozenset(self.servers)
        added_services = list(new_service_set - current_service_set)
        removed_services = list(current_service_set - new_service_set)
        # Add all new services
        for service in added_services:
            service_domain, service_name, service_iface = service
            log.info("Found blob server: %s", service_name)
            d = discovery.resolve(self.reactor, self.browser.service_type,
                                  service_domain, service_name, service_iface)
            d.addCallback(self._blob_server_resolved, service)
    
    def _blob_server_resolved(self, address, service):
        # Log resolution
        host, port = address
        log.debug("Resolved blob server at %s:%i", host, port)
        # Connect to server
        factory = BlobServerFactory(self)
        self.reactor.connectTCP(host, port, factory)
        self.servers[service] = factory
        # Update history
        # Each history item is a (previous_time, session_start) tuple
        #   previous_time is the amount of time accumulated from previous
        #     sessions
        #   session_start is the time at which the current connection started,
        #     or None if there's no current connection.
        now = _get_now()
        server_history = self.history.get(service, [0.0, None])
        if server_history[1] is None:
            # Connection after disconnection
            # Notice that if we're already connected, the history doesn't need
            # to be touched; we just use the old time since there's no break in
            # uptime.
            server_history[1] = now
        self.history[service] = server_history
    
    def fetch_list(self, blob_list):
        """
        Create a blob fetcher for a list of blobs.
        
        :Parameters:
            blob_list : list of UUID
                A list of blob IDs to fetch
        :Returns: A blob fetcher already configured
        :ReturnType: `BlobFetcher`
        """
        return BlobFetcher(self, blob_list)
    
    def get_writer(self):
        """
        Obtains a blob writer for the server list.
        
        :Returns: A new blob writer
        :ReturnType: `BlobWriter`
        """
        return BlobWriter(self)
    
    def drop_server(self, server):
        """
        Drops a server from the list.
        
        Generally, this shouldn't need to be used.  It's mostly for the blob
        servers to inform the list.
        
        :Parameters:
            server : BlobServerFactory
                The server to drop
        """
        now = _get_now()
        for key in list(self.servers):
            if self.servers[key] is server:
                # Log server drop
                service_domain, service_name = key[:2]
                log.info("Dropped blob server: %s", service_name)
                # Drop server
                del self.servers[key]
                # Update history
                server_history = self.history[key]
                server_history[0] += now - server_history[1]
                server_history[1] = None
    
    def get_uptime(self, server):
        """
        Calculate the server's uptime relative to the program's.
        
        :Parameters:
            server : BlobServerFactory
                The blob server to calculate from
        :Raises ValueError: If the server is not in the list
        :Returns: The uptime as a percentage (0.0 to 1.0)
        :ReturnType: float
        """
        # Find server's key
        if isinstance(server, BlobServerFactory):
            for key, value in self.servers.iteritems():
                if value is server:
                    break
            else:
                raise ValueError("Server is not in list: %r" % server)
        else:
            key = server
        # Get server history
        server_history = self.history[key]
        previous_time, session_start = server_history
        # Calculate time
        now = _get_now()
        total_uptime = now - self._start_time
        server_uptime = previous_time
        if session_start is not None:
            server_uptime += now - session_start
        return server_uptime / total_uptime

class BlobFetcher(object):
    """
    Process manager for fetching a series of blobs from the blob servers.
    
    :IVariables:
        has_blobs : twisted.internet.defer.Deferred
            A deferred for when the fetcher figures out whether we have all the
            blobs present
        progress : zfilenet.util.ProgressDeferred
            A deferred for receiving a blob
        finished : twisted.internet.defer.Deferred
            A deferred for completing the fetching
    """
    def __init__(self, server_list, blob_list):
        self.server_list = server_list
        self.servers_tried = set()
        self.blob_index = 0
        self.blob_list = [util.to_uuid(blob_id) for blob_id in blob_list]
        self.blob_map = {}
        # Callbacks
        self.has_blobs = Deferred()
        self.progress = util.ProgressDeferred()
        self.finished = Deferred()
    
    def get_blob_servers(self):
        return self.server_list.servers
    
    def get_next_server(self, blob_id=None):
        if blob_id is None:
            blob_id = self.get_current_blob()
        if self.servers_tried:
            all_servers = frozenset(self.get_blob_servers().iterkeys())
            possible_servers = all_servers - self.servers_tried
            if possible_servers:
                # Still more servers to try, let's have a go.
                key = iter(possible_servers).next()
                blobserver = self.get_blob_servers()[key]
                self.servers_tried.add(key)
                return key, blobserver
            else:
                # No more servers left!
                return None
        else:
            try:
                key, blobserver = self.get_blob_servers().iteritems().next()
            except StopIteration:
                return None
            else:
                self.servers_tried.add(key)
                return key, blobserver
    
    ## MAPPING ##
    
    def start(self):
        """Starts fetching process."""
        self._get_next_list()
    
    def _get_next_list(self):
        # Get next server
        result = self.get_next_server()
        # Determine blobs left
        blobs_left = frozenset(self.blob_list) - frozenset(self.blob_map)
        if blobs_left:
            # Still working
            if result is None:
                # We have no more servers
                self.has_blobs.callback(False)
                self.finished.callback(False)
            else:
                key, server = result
                query_list = [blob_id.hex for blob_id in blobs_left]
                d = server.call('get_blob_list', query_list)
                d.addCallback(self.got_list, key)
        else:
            # Found all we need
            log.debug("Mapped %i blobs", len(self.blob_list))
            self.has_blobs.callback(True)
            self.start_reading()
    
    def got_list(self, blob_list, key):
        for blob_id in blob_list:
            blob_id = util.to_uuid(blob_id)
            self.blob_map[blob_id] = key
        self._get_next_list()
    
    ## READING ##
    
    def get_current_blob(self):
        return self.blob_list[self.blob_index]
    
    def start_reading(self):
        """Starts reading blobs"""
        key = self.blob_map[self.get_current_blob()]
        blobserver = self.get_blob_servers()[key]
        d = blobserver.call('read', self.get_current_blob().hex)
        d.addCallback(self.blob_fetched)
    
    def blob_fetched(self, blob):
        if blob is not None:
            blob = b64decode(blob)
            # This server has the blob!
            self.progress.callback((self.get_current_blob(), blob))
            self.blob_index += 1
            if self.blob_index >= len(self.blob_list):
                # We're finished!  Go home!
                self.finished.callback(True)
                return
        else:
            # Somehow, we didn't get a blob... BORK!!!
            self.finished.callback(False)
        # Find latest blob server
        key = self.blob_map[self.get_current_blob()]
        blobserver = self.get_blob_servers()[key]
        # Contact blob server
        d = blobserver.call('read', self.get_current_blob().hex)
        d.addCallback(self.blob_fetched)

class BlobWriter(object):
    """
    Process manager for writing a series of blobs to the blob servers.
    
    This uses a neural network (`blobann.net`) to choose which servers to
    write to.
    
    .. Tip:: It is desirable to keep a writer around for the duration of
             writing a series of blobs, since it will remember the limits of
             all involved blob servers, thus reducing network overhead as much
             as possible.
    
    :IVariables:
        server_list : `BlobServerList`
            The servers in contact
        limits : dict
            The cached limits of all the servers
    """
    
    def __init__(self, server_list):
        self.server_list = server_list
        self.limits = None
    
    def write_blob(self, blob_id, data):
        """
        Sends a blob to the blob servers.
        
        :Parameters:
            blob_id : UUID-like
                The ID of the blob
            data : str
                The data to send
        """
        # Check to see if we have servers
        if len(self.server_list.servers) == 0:
            raise NoServersError()
        # Create parameters
        info = {'blob_id': util.to_uuid(blob_id), 'data': data}
        info['connections'] = self.server_list.servers.values()
        info['current_server'] = -1
        info['server_count'] = 0
        info['ai_count'] = 0
        info['server_total'] = len(info['connections'])
        # Decide on course of action
        deferred = Deferred()
        if self.limits is None:
            # We don't have any limits cached, so go and find them!
            self.limits = {}
            limit_defer = self._get_all_limits(info['connections'])
            limit_defer.addCallback(self._limits_finished, info, deferred)
        else:
            # We already have some limits, but will it be enough?
            connection_set = frozenset(info['connections'])
            limit_set = frozenset(self.limits)
            missing_limits = connection_set - limit_set
            if not missing_limits:
                # We already have all the limits cached.  Go ahead!
                self._next_server(info, deferred)
            else:
                # Darn, we're missing a few.  No matter, just fetch limits for
                # the ones we don't have.  Pretty smart, eh?
                limit_defer = self._get_all_limits(missing_limits)
                limit_defer.addCallback(self._limits_finished, info, deferred)
        return deferred
    
    def _limits_finished(self, limits, info, deferred):
        self._next_server(info, deferred)
    
    def _get_all_limits(self, connections):
        """Called initially to get the limits of all servers."""
        connections = list(connections)
        log.debug("Getting limits for %i servers", len(connections))
        d = Deferred()
        self._fetch_server_limits(connections, 0, d)
        return d
    
    def _fetch_server_limits(self, connections, index, deferred):
        """Calls the requested server to query limits."""
        server = connections[index]
        d = server.call('get_limits')
        d.addCallback(self._got_server_limits, connections, index, deferred)
    
    def _got_server_limits(self, limits, connections, index, deferred):
        """Called when the limits are received."""
        self.limits[connections[index]] = limits
        index += 1
        if index >= len(connections):
            deferred.callback(self.limits)
        else:
            self._fetch_server_limits(connections, index, deferred)
    
    def _next_server(self, info, deferred):
        # Advance index
        info['current_server'] += 1
        connections = info['connections']
        if info['current_server'] >= len(connections):
            # DONE!!!
            if info['server_count'] == 0:
                # Make sure we have at least one copy
                self._server_write(info, connections[0])
            log.debug("Blob %s: Redundancy %i%% (%i/%i/%i) [%i]",
                      # General statistics
                      info['blob_id'],
                      self._get_redundancy(info) * 100,
                      # Three statistics
                      info['server_count'], info['server_total'],
                      len(connections),
                      # AI Selected
                      info['ai_count'],)
            deferred.callback(info['blob_id'])
            return
        # Get limits
        server = self._get_current_server(info)
        max_blob_size, space_used, quota = self.limits[server]
        # Check if writing is possible
        if quota - space_used < len(info['data']):
            # Nope, we're too full.
            info['server_total'] -= 1
            self.server_list.reactor.callLater(
                0.0, self._next_server, info, deferred)
            return
        # Evaluate factors
        uptime = server.get_uptime()
        space_left = max(1.0 - space_used / quota, 0.0)
        redundancy = self._get_redundancy(info)
        inputs = [uptime, space_left, redundancy]
        should_write = bool(blobann.net.evaluate(inputs)[0])
        if should_write:
            info['ai_count'] += 1
            self._server_write(info)
        # Go to next server
        self.server_list.reactor.callLater(
            0.0, self._next_server, info, deferred)
    
    @staticmethod
    def _get_current_server(info):
        connections = info['connections']
        current_server = info['current_server']
        try:
            return connections[current_server]
        except IndexError:
            log.debug("Trying to get %i (out of %i)", current_server, len(connections))
            raise
    
    @staticmethod
    def _get_redundancy(info):
        return info['server_count'] / info['server_total']
    
    def _server_write(self, info, server=None):
        # Default to current server
        if server is None:
            server = self._get_current_server(info)
        # Get Base64-encoded data (calculating, if necessary)
        try:
            encoded_data = info['encoded_data']
        except KeyError:
            encoded_data = b64encode(info['data'])
            info['encoded_data'] = encoded_data
        # Write to server
        # TODO: Check for errors
        d = server.call('write', info['blob_id'].hex, encoded_data)
        info['server_count'] += 1
        # Update limits
        max_blob_size, space_used, quota = self.limits[server]
        space_used += len(info['data'])
        self.limits[server] = (max_blob_size, space_used, quota)
