#!/usr/bin/env python
#
#   zfilenet/indexserver/index.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Index database manager.

Limits on SQLite3:

* Maximum column size: 1,000,000 bytes (approximately 953 MiB)
* Maximum database size: 2^15 * (2^30 - 1) (approximately 31TiB)
* Recommended database size: <24GiB
* Memory footprint: 256 bytes for every 1 MiB of data (depending on page size)
* Integers: 8 bytes signed=64 bits signed

:Variables:
    log : logging.Logger
        The index log
    DEFAULT_DB_FILE : str
        The name of the default database file, used in `connect`.
"""

from __future__ import absolute_import, with_statement
from base64 import b64decode, b64encode
from datetime import datetime
import logging
import threading
from uuid import UUID, uuid4

from pkg_resources import parse_version
from sqlalchemy import (Table, Column, MetaData, ForeignKey,
                        Integer, String, Text, Binary, Boolean, DateTime)
from sqlalchemy.orm import (backref, mapper, relation, synonym,
                            object_session, sessionmaker)

from . import errors
from .. import db, util

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__docformat__ = 'reStructuredText'
__all__ = ['log',
           'create_session',
           'DEFAULT_DB_FILE',
           'IndexObject',
           'File',
           'Directory',
           'NullObject',
           'Lock',
           'get_root',
           'resolve_path',
           'move',
           'connect',]

log = logging.getLogger('zfilenet.indexserver.index')
create_session = sessionmaker(autoflush=True, transactional=True)

DEFAULT_DB_FILE = 'zfn_index.db'

### TABLES ###

metadata = MetaData()
index_objects_table = Table('index_objects', metadata,
    Column('indexobj_db_id', Integer, primary_key=True),
    Column('indexobj_uuid', String(32), unique=True),
    Column('indexobj_type', String(16), nullable=False),
    Column('indexobj_parent', Integer,
           ForeignKey('index_objects.indexobj_db_id')),
    Column('indexobj_name', String(255)),
    Column('indexobj_created', DateTime),
    Column('indexobj_modified', DateTime),
    Column('indexobj_accessed', DateTime),
)
directories_table = Table('directories', metadata,
    Column('indexobj_db_id', Integer,
           ForeignKey('index_objects.indexobj_db_id'),
           primary_key=True),
)
nulls_table = Table('null_objects', metadata,
    Column('indexobj_db_id', Integer,
           ForeignKey('index_objects.indexobj_db_id'),
           primary_key=True),
)
files_table = Table('files', metadata,
    Column('indexobj_db_id', Integer,
           ForeignKey('index_objects.indexobj_db_id'),
           primary_key=True),
    Column('file_length', Integer),
    Column('file_hash', Binary(256)), # in case we go to SHA256
    # We're using text for our file blob list.  This will work fine on
    # SQLite3.  However, if we ever need to change to MySQL, this will only
    # allow up to 65536 characters.  Whatever.
    Column('file_blob_list', Text),
)
locks_table = Table('locks', metadata,
    Column('lock_id', Integer, primary_key=True),
    Column('lock_name', String(255), unique=True),
    Column('exclusive', Boolean),
)
objects_locks_table = Table('objects_locks', metadata,
    Column('indexobj_db_id', Integer,
           ForeignKey('index_objects.indexobj_db_id')),
    Column('lock_id', Integer, ForeignKey('locks.lock_id')),
)

### MAPPING CLASSES ###

class IndexObject(object):
    """
    Common interface for index objects.
    
    :CVariables:
        is_directory : bool
            Whether the class represents a directory
        is_file : bool
            Whether the class represents a file
        is_file : bool
            Whether the class represents a null
    :IVariables:
        uuid : uuid.UUID
            The object's unique identifier (preserved across moves and renames)
        name : unicode
            The object's name
        created : datetime.datetime
            The object's creation date
        modified : datetime.datetime
            The object's modification date
        accessed : datetime.datetime
            The object's last access date
    """
    is_directory = False
    is_file = False
    is_null = False
    
    @classmethod
    def by_uuid(cls, session, io_uuid):
        io_uuid = util.to_uuid(io_uuid).hex
        return session.query(cls).filter_by(uuid=io_uuid).first()
    
    @classmethod
    def make(cls, session, path, locks=[], replace_nulls=True, commit=True):
        """
        Creates an object at the given path.
        
        If this method is called from a subclass, then the object is of the
        subclass's type.
        
        This method will replace nulls.  If this behavior isn't desired, then
        pass ``False`` to ``replace_nulls``.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            path : path-like
                The path to create at
        :Keywords:
            locks : list
                The locks the client holds.  If this is set to ``None``, then
                locks are ignored.
            replace_nulls : bool
                Whether to replace nulls while creating.  Defaults to ``True``.
            commit : bool
                Whether to commit changes made to the database
        :Raises errors.PathNotFoundError: If the parent path does not exist
        :Raises errors.PathExistsError: If the path already exists
        :Raises errors.WrongTypeError: If the new parent is not a directory
        :Returns: The newly created object
        :ReturnType: `IndexObject`
        """
        # Determine location to create
        path = util.Path(path).convert(absolute=True)
        parent_path, name = path[:-1], path[-1]
        parent = resolve_path(session, parent_path)
        if parent is None:
            raise errors.PathNotFoundError(parent_path, "Parent not found")
        elif not parent.is_directory:
            raise errors.WrongTypeError(parent, "Parent must be a directory")
        elif not parent.can_access(locks):
            raise errors.LockedError(parent_path, "Parent locked")
        # Find child
        child = parent.get_child(name)
        if child is None:
            # Create object
            new_object = cls(name, parent, locks=locks)
            # Save/update the object into the session.
            # The object will probably already be saved if parent != None, since
            # SQLAlchemy cascade is implicitly save-update and merge.
            session.save_or_update(new_object)
        elif child.is_null and replace_nulls:
            # Replace null
            new_object = cls(name)
            child.replace(new_object, locks)
        else:
            raise errors.PathExistsError(path)
        # Commit and return new object
        if commit:
            session.commit()
        return new_object
    
    def __init__(self, name, parent=None, io_uuid=None, locks=[]):
        if io_uuid is None:
            io_uuid = uuid4()
        else:
            io_uuid = util.to_uuid(io_uuid)
        self._uuid = io_uuid.hex
        self.set_name_parent(name, parent, locks)
        # Make sure we set dates afterward, since most operations touch the
        # dates.
        self.created = self.modified = self.accessed = datetime.utcnow()
    
    def __repr__(self):
        return "<%s '%s' (UUID: %s)>" % \
            (type(self).__name__, self.name, self.uuid)
    
    def __str__(self):
        return self.name
    
    ## FILE OPERATIONS ##
    
    def clone(self, parent=None, recursive=True):
        """
        Create a copy of an index object.
        
        The copy will have a different UUID and the file dates will be the same
        as the time of the operation.  This is usually the desired behavior.
        
        The default implementation creates a new index object with the same
        name and parent and calls `touch_access`.
        
        :Parameters:
            parent : `IndexObject`
                The parent of the new clone
            recursive : bool
                Whether the cloning process is recursive.  Ignored in default
                implementation.
        :Returns: A copy of the object
        :ReturnType: `IndexObject`
        """
        self.touch_access()
        return type(self)(self.name, parent)
    
    def delete(self, locks=[], force=False, commit=True):
        """
        Removes the object from the index.
        
        :Parameters:
            locks : list
                The locks the client holds.  If this is set to ``None``, then
                locks are ignored.
        :Keywords:
            force : bool
                Whether to delete children.  Defaults to raising an error on
                directories with children.
            commit : bool
                Whether to commit the session after deleting.  Defaults to
                True.
        :Raises errors.NotSavedError: If the object is not in the database
        """
        session = object_session(self)
        parent = self.get_parent()
        if not self.can_access(locks):
            raise errors.LockedError(self, "Item is locked")
        elif parent is not None and not parent.can_access(locks):
            raise errors.LockedError(parent, "Parent is locked")
        if session is not None:
            session.delete(self)
            if commit:
                session.commit()
        else:
            raise errors.NotSavedError(None)
    
    def lock(self, name=None, exclusive=True):
        """
        Attempts to lock a file.
        
        :Parameters:
            name : str
                The new lock's name (if not given, generate a lock name)
        :Keywords:
            exclusive : bool
                Whether the lock given is the only lock allowed (defaults to
                True)
        :Raises errors.LockedError: If the index object is already locked
        :Returns: The new lock
        :ReturnType: `Lock`
        """
        session = object_session(self)
        # Check for existing lock
        if name is not None:
            new_lock = Lock.by_name(name)
        else:
            new_lock = None
        if new_lock is None:
            new_lock = Lock(name, exclusive=exclusive)
        else:
            exclusive = new_lock.exclusive
        # Check for locks
        for lock in self.locks:
            # Lock logic:
            # If requesting an exclusive lock, require that no other locks
            # are present.  If shared, then ensure there are no exclusive
            # locks.
            if lock.exclusive or exclusive:
                raise errors.LockedError(self)
        # Lock file
        self.locks.append(new_lock)
        # Save new lock (if possible)
        if session is not None:
            session.save_or_update(new_lock)
            session.commit()
        # Return new lock
        return new_lock
    
    def unlock(self, lock):
        """
        Unlocks an index object.
        
        :Parameters:
            lock : `Lock` or str
                Lock to break (or the name of the lock)
        :Raises KeyError: If no lock exists with the given name
        """
        if isinstance(lock, Lock):
            name = lock.name
        else:
            name = lock
        for lock in list(self.locks):
            if lock.name == name:
                lock.destroy()
                break
        else:
            raise KeyError(name)
    
    def get_lock_status(self):
        if self.locks:
            return True
        else:
            return False
    
    def can_access(self, keys, none_ignore=True):
        """
        Decides whether a client can access an index object with its keys.
        
        :Parameters:
            keys : list of str
                Names of `Lock` objects
        :Keywords:
            none_ignore : bool
                Whether passing ``None`` as the keys will immediately grant
                access.  If false, then ``None`` is treated as an empty list.
        :Returns: Whether access is granted
        :ReturnType: bool
        """
        def _to_key(key):
            if isinstance(key, Lock):
                return key.name
            else:
                return key
        # Do we even have locks?
        if not self.get_lock_status():
            return True
        # Check for None input
        if keys is None:
            if none_ignore:
                return True
            else:
                keys = []
        # Do set comparison
        keys = frozenset(_to_key(key) for key in keys)
        locks = frozenset(lock.name for lock in self.locks)
        if self.locks[0].exclusive:
            # Exclusive lock, force need for all keys
            return keys.issuperset(locks)
        else:
            # Shared lock, only require one
            return bool(keys & locks)
    
    ## METADATA ACCESS ##
    
    def stat(self):
        """
        Generates the file statistics table.
        
        This is roughly equivalent to the OS-specific ``stat`` call, which
        provides low-level metadata about a file.  However, this table is more
        high-level and is implemented as a dictionary.
        
        Common keys are:
        
        ==========  =========================================================
         Key         Description
        ==========  =========================================================
        uuid        The UUID of the file (unique across renames and moving)
        type        A short string indicating the type of the file
        created     The time at which the file was created
        modified    The time at which the file was last modified
        accessed    The time at which the file was last accessed
        ==========  =========================================================
        
        :Returns: The stat table
        :ReturnType: dict
        """
        return {'uuid': str(self.uuid),
                'created': util.format_datetime(self.created),
                'modified': util.format_datetime(self.modified),
                'accessed': util.format_datetime(self.accessed),}
    
    def touch(self):
        """
        Update modification date and last access date.
        
        This will unconditionally commit to the database.
        """
        self.modified = self.accessed = datetime.utcnow()
        session = object_session(self)
        if session is not None:
            session.commit()
    
    def touch_access(self):
        """
        Update access date.
        
        This will unconditionally commit to the database.
        """
        self.accessed = datetime.utcnow()
        session = object_session(self)
        if session is not None:
            session.commit()
    
    @property
    def uuid(self):
        return util.to_uuid(self._uuid)
    
    def _get_name(self):
        return self._name
    def _set_name(self, value):
        parent = self.get_parent()
        if util.Path.sep in value or \
           value == util.Path.curdir or \
           value == util.Path.pardir:
            raise errors.BadNameError(value)
        if parent is not None:
            if parent.get_child(value) is not None:
                raise errors.PathExistsError(parent.get_path() + value,
                                             "Name already exists in parent")
            elif not parent.can_access([]):
                raise errors.LockedError(parent, "Parent is locked")
        self._name = value
        self.touch()
    name = property(_get_name, _set_name)
    
    ## HIERARCHY ##
    
    def get_parent(self):
        """
        Retrieves the parent.
        
        :Returns: The parent of the object
        :ReturnType: `Directory`
        """
        return self._parent
    
    def set_parent(self, parent):
        """
        Changes the parent.
        
        :Parameters:
            parent : `Directory`
                The parent to move to
        :Raises errors.WrongTypeError: If the new parent is not a directory
        :Raises errors.PathExistsError: If there is a file with the same name
                                        in the new directory.
        """
        if parent is not None:
            if not parent.is_directory:
                raise errors.WrongTypeError(parent, "Parent must be a directory")
            elif not parent.can_access([]):
                raise errors.LockedError(parent, "Parent is locked")
            child = parent.get_child(self.name)
            if child is not None and child != self:
                raise errors.PathExistsError(parent.get_path() + self.name,
                    "Parent already has a child with that name")
            self._parent = parent
        else:
            self._parent = None
        self.touch()
    
    def set_name_parent(self, name, parent, locks=[], replace_nulls=True):
        """
        Rename and change parents atomically.
        
        This is useful for moving.
        
        :Parameters:
            name : unicode
                The name to change to
            parent : `Directory`
                The new parent
            locks : list
                The locks the client holds.  If this is set to ``None``, then
                locks are ignored.
        :Keywords:
            replace_nulls : bool
                Whether to replace nulls while creating.  Defaults to ``True``.
        :Raises errors.PathExistsError: If the path already exists
        :Raises errors.WrongTypeError: If the new parent is not a directory
        """
        # Check for name
        if util.Path.sep in name or \
           name == util.Path.curdir or \
           name == util.Path.pardir:
            raise errors.BadNameError(name)
        # Check for parent access
        if parent is not None:
            if not parent.is_directory:
                raise errors.WrongTypeError(parent, "Parent must be a directory")
            elif not parent.can_access(locks):
                raise errors.LockedError(parent, "Parent is locked")
        # Check for existing child
        child = parent.get_child(name) if parent is not None else None
        if child is None:
            pass
        elif child == self:
            pass
        elif child.is_null and replace_nulls:
            child.replace(self, locks)
            return
        else:
            raise errors.PathExistsError(parent.get_path() + name,
                "Parent already has a child with that name")
        # Set name and parent
        self._name, self._parent = name, parent
        self.touch()
    
    def get_path(self):
        """
        Gets the path of the object.
        
        :Returns: The object's path
        :ReturnType: `util.Path`
        """
        components = [self.name]
        parent = self.get_parent()
        while parent is not None:
            components.insert(0, parent.name)
            parent = parent.get_parent()
        del components[0] # remove root's name from the path
        return util.Path(components, absolute=True,
                         directory=self.is_directory)

class File(IndexObject):
    """
    An index entry that contains a sequential series of bytes.
    
    :IVariables:
        length : int
            The length of the file (in bytes)
        sha1_hash : str
            The raw binary data of the SHA1 hash
        blob_list : list of UUID
            The file's list of blob IDs
    """
    is_file = True
    
    def __init__(self, *args, **kw):
        super(File, self).__init__(*args, **kw)
        # We could use clear, but that touches the dates
        self.length = 0
        self.sha1_hash = None
        self.blob_list = []
    
    def stat(self):
        stat_table = super(File, self).stat()
        if self.sha1_hash is not None:
            hash_string = util.data2hex(self.sha1_hash)
        else:
            hash_string = None
        stat_table.update(length=self.length,
                          sha1_hash=hash_string,
                          type='file',)
        return stat_table
    
    def clone(self, parent=None, recursive=True):
        new_object = super(File, self).clone(parent, recursive=recursive)
        # Clone attributes
        new_object.length = self.length
        new_object._sha1_hash = self._sha1_hash
        new_object._blob_list = self._blob_list
        # Finish
        return new_object
    
    def _get_sha1_hash(self):
        return str(self._sha1_hash) if self._sha1_hash is not None else None
    def _set_sha1_hash(self, value):
        self._sha1_hash = buffer(value) if value is not None else None
    sha1_hash = property(_get_sha1_hash, _set_sha1_hash)
    
    @staticmethod
    def _encode_blob_list(blob_list):
        return '\n'.join(blob_id.hex for blob_id in blob_list)
    
    @staticmethod
    def _decode_blob_list(blob_list):
        return [util.to_uuid(blob_id) for blob_id in blob_list.split()]
    
    def _get_blob_list(self):
        self.touch_access()
        return self._decode_blob_list(self._blob_list)
    def _set_blob_list(self, value):
        if value is not None:
            blob_list = [util.to_uuid(blob_id) for blob_id in value]
            self._blob_list = self._encode_blob_list(blob_list)
        else:
            self._blob_list = None
        self.touch()
    blob_list = property(_get_blob_list, _set_blob_list)
    
    def add_blob(self, blob_id):
        """
        Appends a blob to the end of the blob list.
        
        This is more efficient than setting `blob_list` when appending.
        
        :Parameters:
            blob_id : UUID-like
                The blob ID to append
        """
        blob_id = util.to_uuid(blob_id).hex
        if self._blob_list:
            self._blob_list += '\n' + blob_id
        else:
            self._blob_list += blob_id
        # Touching commits
        self.touch()
    
    def clear(self):
        """Resets the contents of the file."""
        self.length = 0
        self.sha1_hash = None
        self.blob_list = []
        # Touching commits
        self.touch()

class Directory(IndexObject):
    """A container for other index objects."""
    is_directory = True
    
    def stat(self):
        stat_table = super(Directory, self).stat()
        stat_table.update(type='directory',)
        return stat_table
    
    def clone(self, parent=None, recursive=True):
        new_object = super(Directory, self).clone(parent, recursive=recursive)
        # Copy all children
        # Notice we list-ify the children.  Since a clone MUST do a commit,
        # it messes up the cursor, resulting in very strange errors.
        # Therefore, we must save results and clone from there.
        if recursive:
            for child in list(self.get_children()):
                new_child = child.clone(new_object, recursive=True)
        return new_object
    
    def delete(self, locks=[], force=False, commit=True):
        # Check if we're locked
        parent = self.get_parent()
        if not self.can_access(locks):
            raise errors.LockedError(self, "Directory is locked")
        elif parent is not None and not parent.can_access(locks):
            raise errors.LockedError(parent, "Parent is locked")
        # Check to see if we have children
        try:
            iter(self.get_children()).next()
        except StopIteration:
            has_children = False
        else:
            has_children = True
        # If we have children, require forcing
        if force or not has_children:
            # Delete all children
            for child in list(self.get_children()):
                child.delete(locks, force=force, commit=False)
            # Now delete ourselves
            super(Directory, self).delete(None, force=force, commit=commit)
        else:
            raise PathExistsError(self, "Directory is not empty")
    
    def get_children(self, include_null=True):
        """
        Gets all the children of the directory.
        
        :Keywords:
            include_null : bool
                Whether null resources should be included
        :Returns: A generator of all the child index objects
        :ReturnType: iterator
        """
        self.touch_access()
        if include_null:
            return iter(self._children)
        else:
            return iter(child for child in self._children
                        if not isinstance(child, NullObject))
    
    def get_child(self, name):
        """
        Retrieves a child with a specific name.
        
        :Parameters:
            name : unicode
                The name of the child
        :Returns: The requested child, or ``None`` if not found
        :ReturnType: `IndexObject`
        """
        session = object_session(self)
        if session is not None:
            # We have a session available, let's do a query (faster).
            results = session.query(IndexObject).filter_by(
                _parent=self, name=name)
            return results.first()
        else:
            # No session (not saved yet).  Iterate through children manually.
            for child in self.get_children():
                if child.name == name:
                    return child
            else:
                return None

class NullObject(IndexObject):
    """
    A null entry in a directory.
    
    Typically used for future locking.
    """
    is_null = True
    
    def stat(self):
        stat_table = super(NullObject, self).stat()
        stat_table.update(type='null',)
        return stat_table
    
    def replace(self, other, locks=[]):
        """
        Replaces the null with another object.
        
        :Parameters:
            other : `IndexObject`
                The index object to replace with
            locks : list
                The locks the client holds.  If this is set to ``None``, then
                locks are ignored.
        """
        if not self.can_access(locks):
            raise errors.LockedError(self, "Null is locked")
        session = object_session(self)
        if other not in session:
            session.save(other)
        # Replace with other object
        self.delete(None, commit=False)
        other.set_name_parent(self.name, self.get_parent(), None)
        # Transfer locks
        other.locks += self.locks
        del self.locks[:]
        # Commit changes
        session.commit()

class Lock(object):
    """
    Write lock on an index object.
    
    :IVariables:
        name : str
            The name of the lock
        exclusive : bool
            Whether the lock must be the only lock present
    """
    @classmethod
    def by_name(cls, session, name):
        """
        Retrieves the lock identified by a name.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            name : str
                The lock's name to find
        :Returns: The requested lock, or ``None`` if not found
        :ReturnType: `Lock`
        """
        results = session.query(cls).filter_by(name=name)
        return results.first()
    
    def __init__(self, name=None, exclusive=True):
        if name is None:
            name = "opaquelocktoken:" + str(uuid4())
        self.name = name
        self.exclusive = exclusive
    
    def destroy(self):
        """
        Destroys the lock.
        
        :Raises errors.NotSavedError: If the lock is not in the database
        """
        session = object_session(self)
        if session is not None:
            for obj in self.objects:
                obj.locks.remove(self)
            session.delete(self)
            session.commit()
        else:
            raise errors.NotSavedError(None)
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.destroy()

### MAPPERS ###

mapper(Lock, locks_table,
       properties=dict(
    # Properties
    name=locks_table.c.lock_name,
))
mapper(IndexObject, index_objects_table,
       polymorphic_on=index_objects_table.c.indexobj_type,
       polymorphic_identity='indexobj',
       properties=dict(
    # Properties
    _uuid=index_objects_table.c.indexobj_uuid,
    uuid=synonym('_uuid'),
    _name=index_objects_table.c.indexobj_name,
    name=synonym('_name'),
    _children=relation(IndexObject, backref=backref('_parent',
        remote_side=[index_objects_table.c.indexobj_db_id])),
    created=index_objects_table.c.indexobj_created,
    modified=index_objects_table.c.indexobj_modified,
    accessed=index_objects_table.c.indexobj_accessed,
    locks=relation(Lock, secondary=objects_locks_table,
                   backref='objects'),
))
mapper(File, files_table,
       polymorphic_identity='file', inherits=IndexObject,
       properties=dict(
    # Properties
    length=files_table.c.file_length,
    _sha1_hash=files_table.c.file_hash,
    sha1_hash=synonym('_sha1_hash'),
    _blob_list=files_table.c.file_blob_list,
    blob_list=synonym('_blob_list'),
))
mapper(Directory, directories_table,
       polymorphic_identity='directory', inherits=IndexObject,)
mapper(NullObject, nulls_table,
       polymorphic_identity='null', inherits=IndexObject,)

### HIGH-LEVEL OPERATIONS ###

def get_root(session):
    """
    Retrieves the root of the filesystem hierarchy.
    
    :Parameters:
        session : sqlalchemy.orm.session.Session
            The session to make queries from
    :Returns: The root directory, or ``None`` if not found
    :ReturnType: `Directory`
    """
    root_id = db.DBVariable.get_variable(session, 'root_directory')
    return Directory.by_uuid(session, root_id)

def resolve_path(session, path):
    """
    Finds an `IndexObject` at the indicated path.
    
    :Parameters:
        session : sqlalchemy.orm.session.Session
            The session to make queries from
        path : path-like
            The path to find
    :Returns: The desired node, or ``None`` if not found
    :ReturnType: `IndexObject`
    """
    path = util.Path(path).convert(absolute=True)
    node = get_root(session)
    for component in path:
        try:
            node = node.get_child(component)
        except AttributeError:
            # Not a directory
            return None
        else:
            if node is None:
                # We stopped finding results
                return None
    return node

def move(session, src, dst, locks=[],
         copy=False, recursive=True, allow_directory=True):
    """
    Moves/copies an index object.
    
    :Parameters:
        session : sqlalchemy.orm.session.Session
            The session to make queries from
        src : path-like
            The path to the index object to move from
        dst : path-like
            The path to move the index object to
    :Keywords:
        locks : list
            The locks the client holds.  If this is set to ``None``, then locks
            are ignored.
        copy : bool
            Whether to copy the source (using `IndexObject.clone`) before
            moving to its destination
        recursive : bool
            Whether to copy recursively (only applies to copy operations)
        allow_directory : bool
            Whether to allow the destination to be a directory (e.g. moving
            "a" to "b/" will actually move "a" to "b/a")
    :Raises errors.PathExistsError: If moving the file would overwrite another
                                    file
    :Raises errors.PathNotFoundError: If either the source or the destination
                                      do not exist
    :Raises errors.HierarchyError: If a directory is attempted to be moved into
                                   a subdirectory of itself
    :Raises errors.LockedError: If the source file is being moved while locked
    """
    src_path = util.Path(src).convert(absolute=True)
    dst_path = util.Path(dst).convert(absolute=True)
    src_node = resolve_path(session, src_path)
    dst_node = resolve_path(session, dst_path)
    root = get_root(session)
    if src_node is None or src_node.is_null:
        raise errors.PathNotFoundError(src_path, "Source does not exist")
    elif src_node == root:
        # The crazy user is trying to move root
        raise errors.HierarchyError('/', "Root cannot be moved")
    elif not src_node.can_access(locks) and not copy:
        # Locked source!
        raise errors.LockedError(src_path, "Source locked")
    # Determine destination
    if dst_node is None or dst_node.is_null:
        # See if we're moving to "directory/new_name"
        dst_path, name = dst_path[:-1], dst_path[-1]
        dst_node = resolve_path(session, dst_path)
        if dst_node is None or not dst_node.is_directory:
            raise errors.PathNotFoundError(dst_path,
                "Destination does not exist")
    elif not allow_directory or not dst_node.is_directory:
        # Don't allow overwriting through moving
        raise errors.PathExistsError(dst_path, "Destination exists")
    else:
        # Destination is a directory
        name = src_node.name
    # Prevent cyclic trees
    if src_node.is_directory and \
       list(dst_path[:len(src_path)]) == list(src_path):
        raise errors.HierarchyError(dst_path,
            "Directory cannot move into a subdirectory of itself")
    # Perform move/copy operation
    null = NullObject(name, dst_node, locks=locks)
    with null.lock() as new_lock:
        if copy:
            node = src_node.clone(recursive=recursive)
        else:
            node = src_node
        locks_list = list(locks) if locks is not None else []
        node.set_name_parent(name, dst_node, locks_list + [new_lock])
    session.commit()

def _on_new(engine):
    from .. import indexserver
    log.debug("Creating new database...")
    session = create_session(bind=engine)
    db.DBVariable.set_variable(session, 'version', indexserver.__version__)
    new_root = Directory('root')
    session.save(new_root)
    db.DBVariable.set_variable(session, 'root_directory', new_root.uuid.hex)
    log.debug("Root UUID: %s", new_root.uuid)
    session.commit()

def connect(db_file=DEFAULT_DB_FILE):
    """
    Connects to an index database file.
    
    **Always use this method when connecting to an index database!**  This
    method knows the correct procedure for initializing a filesystem.
    
    :Parameters:
        db_file : str
            The path to the database file.  If no path is given, then the
            default name is used.
    :Returns: The database engine
    :ReturnType: sqlalchemy.engine.Engine
    """
    # Connect to database
    engine = db.connect_sqlite3(metadata, db_file, _on_new)
    # Determine version
    session = create_session(bind=engine)
    db_version = db.DBVariable.get_variable(session, 'version')
    del session
    # Minimum and maximum versions allowed
    # NOTE: Change for each release
    min_version = parse_version('0.2.0a')
    max_version = parse_version('0.3.0a')
    # Do version check
    if not min_version <= parse_version(db_version) < max_version \
       and db_version != 'dev':
        raise errors.WrongVersionError(None, db_version)
    return engine
