#!/usr/bin/env python
#
#   zfilenet/tests/test_index.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Testing module for the index"""

import unittest

from zfilenet.indexserver import errors
from zfilenet.indexserver.index import (connect, create_session,
                                        IndexObject, NullObject,
                                        File, Directory,
                                        get_root, resolve_path, move)

__author__ = 'Ross Light'
__date__ = 'July 20, 2008'
__all__ = ['DataTest']

class DataTest(unittest.TestCase):
    def setUp(self):
        self.engine = connect(':memory:')
        self.session = create_session(bind=self.engine)
        self.root = get_root(self.session)
    
    def tearDown(self):
        self.session.close()
    
    def test_clone(self):
        """File cloning should have the same information"""
        src_file = File("lumberjack.txt")
        clone = src_file.clone()
        assert clone.uuid != src_file.uuid, "UUIDs should not match"
        assert clone.name == src_file.name, "Names do not match"
        assert clone.is_file, "Clone changed types"
    
    def test_directory_stat(self):
        """Directory stat test"""
        from zfilenet.util import to_uuid, format_datetime
        self.root.touch()
        self.root.touch_access()
        result = self.root.stat()
        assert result['type'] == 'directory', "Type is wrong"
        assert result['uuid'] == str(self.root.uuid), "UUID is wrong"
        assert result['created'] == format_datetime(self.root.created), \
            "Creation date is wrong"
        assert result['modified'] == format_datetime(self.root.modified), \
            "Modification date is wrong"
        assert result['accessed'] == format_datetime(self.root.accessed), \
            "Access date is wrong"
    
    def test_file_stat(self):
        """File stat test"""
        from zfilenet.util import to_uuid, format_datetime, data2hex
        new_file = File('blab')
        new_file.touch()
        new_file.touch_access()
        result = new_file.stat()
        assert result['type'] == 'file', "Type is wrong"
        assert result['uuid'] == str(new_file.uuid), "UUID is wrong"
        assert result['created'] == format_datetime(new_file.created), \
            "Creation date is wrong"
        assert result['modified'] == format_datetime(new_file.modified), \
            "Modification date is wrong"
        assert result['accessed'] == format_datetime(new_file.accessed), \
            "Access date is wrong"
        assert result['length'] == 0, "Length is wrong"
        assert result['sha1_hash'] is None, "Hash should not be present"
    
    def test_path(self):
        """File path test"""
        new_directory = Directory.make(self.session, '/spam')
        assert new_directory.get_path() == '/spam/', "Wrong directory path"
        new_file = File.make(self.session, '/spam/eggs.txt')
        assert new_file.get_path() == '/spam/eggs.txt', "Wrong file path"
    
    def test_resolve_path(self):
        """Path resolution test"""
        assert resolve_path(self.session, '/') is self.root, \
            "Root doesn't resolve"
        new_directory = Directory.make(self.session, '/spam')
        assert resolve_path(self.session, '/spam') is new_directory, \
            "New directory doesn't resolve"
        new_file = File.make(self.session, '/spam/eggs.txt')
        assert resolve_path(self.session, '/spam/eggs.txt') is new_file, \
            "New file doesn't resolve"
    
    def test_exclusive_locking(self):
        """Exclusive locking test"""
        my_file = File.make(self.session, '/spam.txt')
        # Make sure we're unlocked
        assert my_file.can_access([]), "Unlocked file should be accessible"
        assert not my_file.get_lock_status(), \
            "Unlocked file should have lock status = False"
        # Test exclusive lock
        lock1 = my_file.lock(exclusive=True)
        assert not my_file.can_access([]), \
            "Locked file shouldn't be accessible"
        assert my_file.can_access([lock1]), "Key does not unlock file"
        # Test making other locks (should fail)
        try:
            my_file.lock(exclusive=True)
        except errors.LockedError:
            pass
        else:
            raise AssertionError("Taking second exclusive lock succeeded")
        try:
            my_file.lock(exclusive=False)
        except errors.LockedError:
            pass
        else:
            raise AssertionError("Taking shared lock succeeded")
        # Test unlocking
        my_file.unlock(lock1)
        assert my_file.can_access([]), "File should be unlocked"
        assert not my_file.get_lock_status(), \
            "File should have lock status = False"
    
    def test_shared_locking(self):
        """Shared locking test"""
        my_file = File.make(self.session, '/spam.txt')
        # Make sure we're unlocked
        assert my_file.can_access([]), "Unlocked file should be accessible"
        assert not my_file.get_lock_status(), \
            "Unlocked file should have lock status = False"
        # Test exclusive lock
        lock1 = my_file.lock(exclusive=False)
        assert not my_file.can_access([]), \
            "Locked file shouldn't be accessible"
        assert my_file.can_access([lock1]), "Key does not unlock file"
        # Test making other locks (exclusive should fail)
        try:
            my_file.lock(exclusive=True)
        except errors.LockedError:
            pass
        else:
            raise AssertionError("Taking exclusive lock succeeded")
        try:
            lock2 = my_file.lock(exclusive=False)
        except errors.LockedError:
            raise AssertionError("Taking second shared lock failed")
        # Test unlocking
        my_file.unlock(lock2)
        my_file.unlock(lock1)
        assert my_file.can_access([]), "File should be unlocked"
        assert not my_file.get_lock_status(), \
            "File should have lock status = False"
    
    def test_null_replace(self):
        """Replacing nulls should destroy original null"""
        my_null = NullObject.make(self.session, '/null')
        assert my_null.is_null, "Null is not null"
        new_file = File('spam')
        creation = new_file.created
        my_null.replace(new_file)
        assert new_file.created == creation, "Creation date wrong"
        assert new_file.name == 'null', "Name not changed"
        result = resolve_path(self.session, '/null')
        assert result is new_file, "Null not replaced"
    
    def test_locked_delete(self):
        """Deleting a locked file should fail"""
        my_file = File.make(self.session, '/spam.eggs')
        lock = my_file.lock()
        # Try to fail
        try:
            my_file.delete()
        except errors.LockedError:
            pass
        else:
            raise AssertionError("File allowed deletion")
        # Try to succeed
        try:
            my_file.delete([lock])
        except errors.LockedError:
            raise AssertionError("File should've allowed delete")
    
    def test_move(self):
        """File move test"""
        file1 = File.make(self.session, '/spam')
        lock1 = file1.lock()
        null = NullObject.make(self.session, '/bacon')
        lock2 = null.lock()
        # Try to move without key
        try:
            move(self.session, '/spam', '/eggs')
        except errors.LockedError:
            pass
        else:
            raise AssertionError("Locked file allowed move")
        # Try to copy
        try:
            move(self.session, '/spam', '/eggs', copy=True)
        except errors.LockedError:
            raise AssertionError("Locked file denied copy")
        # Check for copy similarity
        file2 = resolve_path(self.session, '/eggs')
        assert file1.uuid != file2.uuid, "Same UUID"
        assert file2.is_file, "Copy is not a file"
        # Try to move without null lock
        try:
            move(self.session, '/spam', '/bacon', [lock1])
        except errors.LockedError:
            pass
        else:
            raise AssertionError("Locked null allowed replacement")
        # Try to move with both locks
        try:
            move(self.session, '/spam', '/ham', [lock1, lock2])
        except errors.LockedError:
            raise AssertionError("Moving is overly locked")
        # Check for move change
        file3 = resolve_path(self.session, '/ham')
        assert file3 is file1, "Move split file"
        assert file1.name == 'ham', "Name has not changed"
        # Destroy locks
        lock2.destroy()
        lock1.destroy()
